@@ -13,6 +13,7 @@ from typing import TYPE_CHECKING, Any, Callable, Iterable, NamedTuple, Sequence,
 
 from hatchling.__about__ import __version__
 from hatchling.builders.config import BuilderConfig
+from hatchling.builders.constants import EDITABLES_REQUIREMENT
 from hatchling.builders.plugin.interface import BuilderInterface
 from hatchling.builders.utils import (
     format_file_hash,
@@ -32,8 +33,6 @@ if TYPE_CHECKING:
     from hatchling.builders.plugin.interface import IncludedFile
 
 
-EDITABLES_MINIMUM_VERSION = '0.3'
-
 TIME_TUPLE = Tuple[int, int, int, int, int, int]
 
 
@@ -124,6 +123,10 @@ class WheelArchive:
         relative_path = f'{self.metadata_directory}/{normalize_archive_path(relative_path)}'
         return self.write_file(relative_path, contents)
 
+    def write_shared_script(self, relative_path: str, contents: str | bytes) -> tuple[str, str, str]:
+        relative_path = f'{self.shared_data_directory}/scripts/{normalize_archive_path(relative_path)}'
+        return self.write_file(relative_path, contents)
+
     def add_shared_file(self, shared_file: IncludedFile) -> tuple[str, str, str]:
         shared_file.distribution_path = f'{self.shared_data_directory}/data/{shared_file.distribution_path}'
         return self.add_file(shared_file)
@@ -164,6 +167,7 @@ class WheelBuilderConfig(BuilderConfig):
 
         self.__core_metadata_constructor: Callable[..., str] | None = None
         self.__shared_data: dict[str, str] | None = None
+        self.__shared_scripts: dict[str, str] | None = None
         self.__extra_metadata: dict[str, str] | None = None
         self.__strict_naming: bool | None = None
         self.__macos_max_compat: bool | None = None
@@ -178,6 +182,7 @@ class WheelBuilderConfig(BuilderConfig):
         if include or packages or only_include:
             return FileSelectionOptions(include, exclude, packages, only_include)
 
+        project_names: set[str] = set()
         for project_name in (
             self.builder.normalize_file_name_component(self.builder.metadata.core.raw_name),
             self.builder.normalize_file_name_component(self.builder.metadata.core.name),
@@ -201,19 +206,24 @@ class WheelBuilderConfig(BuilderConfig):
                 relative_path = os.path.relpath(possible_namespace_packages[0], self.root)
                 namespace = relative_path.split(os.sep)[0]
                 return FileSelectionOptions([], exclude, [namespace], [])
+            project_names.add(project_name)
 
         if self.bypass_selection or self.build_artifact_spec is not None or self.get_force_include():
             self.set_exclude_all()
             return FileSelectionOptions([], exclude, [], [])
 
+        project_names_text = ' or '.join(sorted(project_names))
         message = (
-            'Unable to determine which files to ship inside the wheel using the following heuristics: '
-            'https://hatch.pypa.io/latest/plugins/builder/wheel/#default-file-selection\n\nAt least one '
-            'file selection option must be defined in the `tool.hatch.build.targets.wheel` table, see: '
-            'https://hatch.pypa.io/latest/config/build/\n\nAs an example, if you intend to ship a '
-            'directory named `foo` that resides within a `src` directory located at the root of your '
-            'project, you can define the following:\n\n[tool.hatch.build.targets.wheel]\n'
-            'packages = ["src/foo"]'
+            f'Unable to determine which files to ship inside the wheel using the following heuristics: '
+            f'https://hatch.pypa.io/latest/plugins/builder/wheel/#default-file-selection\n\n'
+            f'The most likely cause of this is that there is no directory that matches the name of your '
+            f'project ({project_names_text}).\n\n'
+            f'At least one file selection option must be defined in the `tool.hatch.build.targets.wheel` '
+            f'table, see: https://hatch.pypa.io/latest/config/build/\n\n'
+            f'As an example, if you intend to ship a directory named `foo` that resides within a `src` '
+            f'directory located at the root of your project, you can define the following:\n\n'
+            f'[tool.hatch.build.targets.wheel]\n'
+            f'packages = ["src/foo"]'
         )
         raise ValueError(message)
 
@@ -284,6 +294,40 @@ class WheelBuilderConfig(BuilderConfig):
 
         return self.__shared_data
 
+    @property
+    def shared_scripts(self) -> dict[str, str]:
+        if self.__shared_scripts is None:
+            shared_scripts = self.target_config.get('shared-scripts', {})
+            if not isinstance(shared_scripts, dict):
+                message = f'Field `tool.hatch.build.targets.{self.plugin_name}.shared-scripts` must be a mapping'
+                raise TypeError(message)
+
+            for i, (source, relative_path) in enumerate(shared_scripts.items(), 1):
+                if not source:
+                    message = (
+                        f'Source #{i} in field `tool.hatch.build.targets.{self.plugin_name}.shared-scripts` '
+                        f'cannot be an empty string'
+                    )
+                    raise ValueError(message)
+
+                if not isinstance(relative_path, str):
+                    message = (
+                        f'Path for source `{source}` in field '
+                        f'`tool.hatch.build.targets.{self.plugin_name}.shared-scripts` must be a string'
+                    )
+                    raise TypeError(message)
+
+                if not relative_path:
+                    message = (
+                        f'Path for source `{source}` in field '
+                        f'`tool.hatch.build.targets.{self.plugin_name}.shared-scripts` cannot be an empty string'
+                    )
+                    raise ValueError(message)
+
+            self.__shared_scripts = normalize_inclusion_map(shared_scripts, self.root)
+
+        return self.__shared_scripts
+
     @property
     def extra_metadata(self) -> dict[str, str]:
         if self.__extra_metadata is None:
@@ -459,7 +503,7 @@ class WheelBuilder(BuilderInterface):
                     try:
                         exposed_packages[distribution_module] = os.path.join(
                             self.root,
-                            f'{relative_path[:relative_path.index(distribution_path)]}{distribution_module}',
+                            f'{relative_path[: relative_path.index(distribution_path)]}{distribution_module}',
                         )
                     except ValueError:
                         message = (
@@ -494,7 +538,7 @@ class WheelBuilder(BuilderInterface):
             for raw_dependency in editable_project.dependencies():
                 dependency = raw_dependency
                 if dependency == 'editables':
-                    dependency += f'~={EDITABLES_MINIMUM_VERSION}'
+                    dependency = EDITABLES_REQUIREMENT
                 else:  # no cov
                     pass
 
@@ -542,6 +586,7 @@ class WheelBuilder(BuilderInterface):
         self, archive: WheelArchive, records: RecordFile, build_data: dict[str, Any], extra_dependencies: Sequence[str]
     ) -> None:
         self.add_shared_data(archive, records)
+        self.add_shared_scripts(archive, records)
 
         # Ensure metadata is written last, see https://peps.python.org/pep-0427/#recommended-archiver-features
         self.write_metadata(archive, records, build_data, extra_dependencies=extra_dependencies)
@@ -551,6 +596,35 @@ class WheelBuilder(BuilderInterface):
             record = archive.add_shared_file(shared_file)
             records.write(record)
 
+    def add_shared_scripts(self, archive: WheelArchive, records: RecordFile) -> None:
+        import re
+        from io import BytesIO
+
+        # https://packaging.python.org/en/latest/specifications/binary-distribution-format/#recommended-installer-features
+        shebang = re.compile(rb'^#!.*(?:pythonw?|pypyw?)[0-9.]*(.*)', flags=re.DOTALL)
+
+        for shared_script in self.recurse_explicit_files(self.config.shared_scripts):
+            with open(shared_script.path, 'rb') as f:
+                content = BytesIO()
+                for line in f:
+                    # Ignore leading blank lines
+                    if not line.strip():
+                        continue
+
+                    match = shebang.match(line)
+                    if match is None:
+                        content.write(line)
+                    else:
+                        content.write(b'#!python')
+                        if remaining := match.group(1):
+                            content.write(remaining)
+
+                    content.write(f.read())
+                    break
+
+            record = archive.write_shared_script(shared_script.distribution_path, content.getvalue())
+            records.write(record)
+
     def write_metadata(
         self,
         archive: WheelArchive,
@@ -644,13 +718,30 @@ Root-Is-Purelib: {'true' if build_data['pure_python'] else 'false'}
         return metadata_file.lstrip()
 
     def get_default_tag(self) -> str:
+        known_major_versions = list(get_known_python_major_versions())
+        max_version_part = 100
         supported_python_versions = []
-        for major_version in get_known_python_major_versions():
-            for minor_version in range(100):
-                if self.metadata.core.python_constraint.contains(f'{major_version}.{minor_version}'):
+        for major_version in known_major_versions:
+            for minor_version in range(max_version_part):
+                # Try an artificially high patch version to account for common cases like `>=3.11.4` or `>=3.10,<3.11`
+                if self.metadata.core.python_constraint.contains(f'{major_version}.{minor_version}.{max_version_part}'):
                     supported_python_versions.append(f'py{major_version}')
                     break
 
+        # Slow path, try all permutations to account for narrow support ranges like `<=3.11.4`
+        if not supported_python_versions:
+            for major_version in known_major_versions:
+                for minor_version in range(max_version_part):
+                    for patch_version in range(max_version_part):
+                        if self.metadata.core.python_constraint.contains(
+                            f'{major_version}.{minor_version}.{patch_version}'
+                        ):
+                            supported_python_versions.append(f'py{major_version}')
+                            break
+                    else:
+                        continue
+                    break
+
         return f'{".".join(supported_python_versions)}-none-any'
 
     def get_best_matching_tag(self) -> str:
@@ -672,7 +763,7 @@ Root-Is-Purelib: {'true' if build_data['pure_python'] else 'false'}
                     plat = tag_parts[2]
                     current_arch = platform.mac_ver()[2]
                     new_arch = 'universal2' if set(archs) == {'x86_64', 'arm64'} else archs[0]
-                    tag_parts[2] = f'{plat[:plat.rfind(current_arch)]}{new_arch}'
+                    tag_parts[2] = f'{plat[: plat.rfind(current_arch)]}{new_arch}'
 
             if self.config.macos_max_compat:
                 import re
